@@ -72,34 +72,42 @@ class Attention(nn.Module):
         return self.to_out(out)
 
 class SpatialGatingUnit(nn.Module):
-    def __init__(self, dim, dim_seq, causal = False, act = nn.Identity(), init_eps = 1e-3, use_circulant_matrix = False):
+    def __init__(
+        self,
+        dim,
+        dim_seq,
+        causal = False,
+        act = nn.Identity(),
+        heads = 1,
+        init_eps = 1e-3,
+        use_circulant_matrix = False
+    ):
         super().__init__()
         dim_out = dim // 2
+        self.heads = heads
         self.causal = causal
-
         self.norm = nn.LayerNorm(dim_out)
-        self.proj = nn.Conv1d(dim_seq, dim_seq, 1)
 
         self.act = act
 
         # parameters
 
         if use_circulant_matrix:
-            self.circulant_pos_x = nn.Parameter(torch.ones((dim_seq,)))
-            self.circulant_pos_y = nn.Parameter(torch.ones((dim_seq,)))
+            self.circulant_pos_x = nn.Parameter(torch.ones(heads, dim_seq))
+            self.circulant_pos_y = nn.Parameter(torch.ones(heads, dim_seq))
 
         self.use_circulant_matrix = use_circulant_matrix
-        shape = (dim_seq,) if use_circulant_matrix else (dim_seq, dim_seq)
+        shape = (heads, dim_seq,) if use_circulant_matrix else (heads, dim_seq, dim_seq)
         weight = torch.zeros(shape)
 
         self.weight = nn.Parameter(weight)
         init_eps /= dim_seq
         nn.init.uniform_(self.weight, -init_eps, init_eps)
 
-        self.bias = nn.Parameter(torch.ones(dim_seq))
+        self.bias = nn.Parameter(torch.ones(heads, dim_seq))
 
     def forward(self, x, gate_res = None):
-        device, n = x.device, x.shape[1]
+        device, n, h = x.device, x.shape[1], self.heads
 
         res, gate = x.chunk(2, dim = -1)
         gate = self.norm(gate)
@@ -107,22 +115,31 @@ class SpatialGatingUnit(nn.Module):
         weight, bias = self.weight, self.bias
 
         if self.use_circulant_matrix:
+            # build the circulant matrix
+
             dim_seq = weight.shape[-1]
             weight = F.pad(weight, (0, dim_seq), value = 0)
             weight = repeat(weight, '... n -> ... (r n)', r = dim_seq)
-            weight = weight[:-dim_seq].reshape(dim_seq, 2 * dim_seq - 1)
-            weight = weight[:, (dim_seq - 1):]
+            weight = weight[:, :-dim_seq].reshape(h, dim_seq, 2 * dim_seq - 1)
+            weight = weight[:, :, (dim_seq - 1):]
+
+            # give circulant matrix absolute position awareness
 
             pos_x, pos_y = self.circulant_pos_x, self.circulant_pos_y
-            weight = weight * rearrange(pos_x, 'i -> i ()') * rearrange(pos_y, 'j -> () j')
+            weight = weight * rearrange(pos_x, 'h i -> h i ()') * rearrange(pos_y, 'h j -> h () j')
 
         if self.causal:
-            weight, bias = weight[:n, :n], bias[:n]
-            mask = torch.ones(weight.shape[:2], device = device).triu_(1).bool()
+            weight, bias = weight[:, :n, :n], bias[:, :n]
+            mask = torch.ones(weight.shape[-2:], device = device).triu_(1).bool()
+            mask = rearrange(mask, 'i j -> () i j')
             weight = weight.masked_fill(mask, 0.)
 
-        gate = einsum('b n d, m n -> b m d', gate, weight)
-        gate = gate + rearrange(bias, 'n -> () n ()')
+        gate = rearrange(gate, 'b n (h d) -> b h n d', h = h)
+
+        gate = einsum('b h n d, h m n -> b h m d', gate, weight)
+        gate = gate + rearrange(bias, 'h n -> () h n ()')
+
+        gate = rearrange(gate, 'b h n d -> b n (h d)')
 
         if exists(gate_res):
             gate = gate + gate_res
@@ -136,6 +153,7 @@ class gMLPBlock(nn.Module):
         dim,
         dim_ff,
         seq_len,
+        heads = 1,
         attn_dim = None,
         causal = False,
         act = nn.Identity(),
@@ -149,7 +167,7 @@ class gMLPBlock(nn.Module):
 
         self.attn = Attention(dim, dim_ff // 2, attn_dim, causal) if exists(attn_dim) else None
 
-        self.sgu = SpatialGatingUnit(dim_ff, seq_len, causal, act, use_circulant_matrix = use_circulant_matrix)
+        self.sgu = SpatialGatingUnit(dim_ff, seq_len, causal, act, heads, use_circulant_matrix = use_circulant_matrix)
         self.proj_out = nn.Linear(dim_ff // 2, dim)
 
     def forward(self, x):
@@ -170,6 +188,7 @@ class gMLP(nn.Module):
         dim,
         depth,
         seq_len,
+        heads = 1,
         ff_mult = 4,
         attn_dim = None,
         prob_survival = 1.,
@@ -178,13 +197,15 @@ class gMLP(nn.Module):
         use_circulant_matrix = False
     ):
         super().__init__()
+        assert (dim % heads) == 0, 'dimension must be divisible by number of heads'
+
         dim_ff = dim * ff_mult
         self.seq_len = seq_len
         self.prob_survival = prob_survival
 
         self.to_embed = nn.Embedding(num_tokens, dim) if exists(num_tokens) else nn.Identity()
 
-        self.layers = nn.ModuleList([Residual(PreNorm(dim, gMLPBlock(dim = dim, dim_ff = dim_ff, seq_len = seq_len, attn_dim = attn_dim, causal = causal, act = act, use_circulant_matrix = use_circulant_matrix))) for i in range(depth)])
+        self.layers = nn.ModuleList([Residual(PreNorm(dim, gMLPBlock(dim = dim, heads = heads, dim_ff = dim_ff, seq_len = seq_len, attn_dim = attn_dim, causal = causal, act = act, use_circulant_matrix = use_circulant_matrix))) for i in range(depth)])
 
         self.to_logits = nn.Sequential(
             nn.LayerNorm(dim),
@@ -206,12 +227,15 @@ class gMLPVision(nn.Module):
         num_classes,
         dim,
         depth,
+        heads = 1,
         ff_mult = 4,
         channels = 3,
         attn_dim = None,
         prob_survival = 1.
     ):
         super().__init__()
+        assert (dim % heads) == 0, 'dimension must be divisible by number of heads'
+
         image_height, image_width = pair(image_size)
         patch_height, patch_width = pair(patch_size)
         assert (image_height % patch_height) == 0 and (image_width % patch_width) == 0, 'image height and width must be divisible by patch size'
@@ -226,7 +250,7 @@ class gMLPVision(nn.Module):
 
         self.prob_survival = prob_survival
 
-        self.layers = nn.ModuleList([Residual(PreNorm(dim, gMLPBlock(dim = dim, dim_ff = dim_ff, seq_len = num_patches, attn_dim = attn_dim))) for i in range(depth)])
+        self.layers = nn.ModuleList([Residual(PreNorm(dim, gMLPBlock(dim = dim, heads = heads, dim_ff = dim_ff, seq_len = num_patches, attn_dim = attn_dim))) for i in range(depth)])
 
         self.to_logits = nn.Sequential(
             nn.LayerNorm(dim),
